package org.nm.challenge.wave;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;

import javax.swing.SwingUtilities;

import org.jfree.ui.RefineryUtilities;
import org.nm.challenge.arrays.Array2D;
import org.nm.challenge.display.CombinedDisplayFrame;
import org.nm.challenge.display.IDisplay;

public class Wave {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		int debug = 0;
		int nThreads = 2;
		int nIters = 1000;
		int displayIters = 10;
		long displayWait = 000;
		boolean wavePhysics = true;
		Initialization initType = Initialization.LINEAR_X;
		final int size_y = 500;
		final int size_x = 200;
		Wave wave = new Wave(size_y, size_x, initType, wavePhysics,
				nThreads, nIters, debug, displayIters, displayWait);
		
		// Start the simulation.
		// This creates the threads
		// and starts them running.
		wave.start();
	}
	
	// CyclicBarrier is used to synchronize threads.
	// It's like saying to all of the threads, "Everyone
	// meet up here before you go off on your own thing."
	
	private final CyclicBarrier initializationBarrier;
	private final CyclicBarrier iterationBarrier;

	private final int nThreads;
	private final int nIterations;
	private final List<IWaveWorker> workers;
	private double totalH;
	private int iteration = 0;
	private final int displayIters;
	private int debug;
	private final IDisplay display;
	private final FullWaveProblemSetup probSetup;
	private final long displayWait;
	private final Initialization initType;
	
	public Wave(int size_y, int size_x, Initialization initType,
			boolean wavePhysics, int nThreads, int nIterations, int debug,
			int displayIters, long displayWait) {
		System.out.println(nThreads + " threads, " + nIterations + " iterations.");
		this.displayWait = displayWait;
		this.nThreads = nThreads;
		this.nIterations = nIterations;
		this.displayIters = displayIters;
		this.initType = initType;

		// When all of the workers synchronize on this
		// barrier the iterationCompleted() method will
		// be called *before* the workers are allowed to
		// continue.
		this.iterationBarrier = 
			new CyclicBarrier(nThreads,	new Runnable() {
				public void run() {
					iterationCompleted();
				}
			});

		// When all of the workers synchronize on this
		// barrier the initializationCompleted() method will
		// be called *before* the workers are allowed to
		// continue.
		this.initializationBarrier = 
			new CyclicBarrier(nThreads,	new Runnable() {
				public void run() {
					initializationCompleted();
				}
			});

		this.workers = new ArrayList<IWaveWorker>(nThreads);
		this.debug = debug;
		this.display = createDisplay();
		probSetup = new FullWaveProblemSetup(size_y, size_x, wavePhysics);
	}

	public int getNIterations() {
		return nIterations;
	}

	public int getNThreads() {
		return nThreads;
	}

	public synchronized double getTotalH() {
		return totalH;
	}

	public void workerInitCompleted(IWaveWorker worker) {
		// All of the wave worker threads will call this
		// after they have completed their initialization
		// of the arrays.  When the last one calls this method
		// the initializationBarrier will automatically
		// call initializationCompleted() before letting the worker
		// threads go on their way.  The initializationCompleted()
		// method will update the display and calculate the total H.
		try {
			initializationBarrier.await();
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (BrokenBarrierException e) {
			e.printStackTrace();
		}
	}
	
	public void workerIterationCompleted(IWaveWorker worker) {
		// All of the wave worker threads will call this
		// after they have completed an iteration.
		// When the last one calls this method
		// the iterationBarrier will automatically
		// call iterationCompleted() before letting the worker
		// threads go on their way.  The iterationCompleted()
		// method will update the display and calculate the total H.
		try {
			iterationBarrier.await();
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (BrokenBarrierException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Start each worker in its own thread.
	 */
	public void start() {
		long time = System.currentTimeMillis();
		System.out.println("creating workers");
		createWorkers();
		// start each worker in its own thread
		System.out.println("starting workers");
		ArrayList<Thread> threads = new ArrayList<Thread>(nThreads);
		for (IWaveWorker worker : workers) {
			final Thread thread = new Thread(worker);
			threads.add(thread);
			
			// When start() is called on a thread
			// a new thread is created and the
			// worker's run() method is called.
			thread.start();
		}
		
		for (Thread thread : threads) {
			try {
				// Wait for all of the threads to end
				// execution.
				thread.join();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		updateDisplay();
		long endTime = System.currentTimeMillis();
		System.out.println("done.  Total time: " + (endTime-time));
	}
	
	private void calcTotalH() {
		double totalH = 0;
		for (IWaveWorker worker : workers) {
			totalH += worker.getLocalH();
		}
		if (debug > 0)
			System.out.println("iteration: " + iteration  + " totalH: " + totalH);
	}

	private IDisplay createDisplay() {
		IDisplay retval = null;
		if (displayIters > 0) {
			//JFreeChartDisplayFrame display = new JFreeChartDisplayFrame();
			//Display2DFrame display = new Display2DFrame();
			CombinedDisplayFrame display = new CombinedDisplayFrame();

			display.pack();
			display.setSize(display.getPreferredSize());
			RefineryUtilities.centerFrameOnScreen(display);
			display.setVisible(true);
			retval = display;
		}
		return retval;
	}

	private void createWorkers() {
		for (int rank=0; rank<nThreads; ++rank) {
			workers.add(new WaveWorker(this, initType, probSetup, rank, nThreads, debug));
		}
	}

	private void initializationCompleted() {
		System.out.println("in initializationCompleted()");
		calcTotalH();
		updateDisplay();
		System.out.println("leaving initializationCompleted()");
	}

	private void iterationCompleted() {
		++iteration;
		calcTotalH();
		probSetup.swapArrays();
		if (displayIters > 0 && (iteration-1) % displayIters == 0) {
			updateDisplay();
		}
	}

	private void updateDisplay() {
		if (display == null) {
			return;
		}
		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				public void run() {
					display.update(iteration, new Array2D(probSetup.getH()));
				}
			});
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}
		try {
			Thread.sleep(displayWait);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

}
